from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import AgentOnTileGoal
from ...core.grid import four_rooms, sample_coordinates, sample_direction
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

# goals and rules are hardcoded for minigrid envs
_goal_encoding = AgentOnTileGoal(tile=TILES_REGISTRY[Tiles.GOAL, Colors.GREEN]).encode()
_rule_encoding = EmptyRule().encode()[None, ...]


class FourRooms(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=19, width=19)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=100)
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, *keys = jax.random.split(key, num=4)

        grid = four_rooms(params.height, params.width)

        black_floor = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]
        # sampling doors positions
        doors_offsets = jax.random.randint(keys[0], shape=(4,), minval=1, maxval=params.height // 2)
        grid = grid.at[params.height // 2, doors_offsets[0]].set(black_floor)
        grid = grid.at[params.height // 2, params.width // 2 + doors_offsets[1]].set(black_floor)
        grid = grid.at[doors_offsets[2], params.width // 2].set(black_floor)
        grid = grid.at[params.height // 2 + doors_offsets[3], params.width // 2].set(black_floor)

        # sampling agent and goal positions
        goal_coords, agent_coords = sample_coordinates(keys[1], grid, num=2)
        grid = grid.at[goal_coords[0], goal_coords[1]].set(TILES_REGISTRY[Tiles.GOAL, Colors.GREEN])

        agent = AgentState(position=agent_coords, direction=sample_direction(keys[2]))
        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

import importlib
from dataclasses import dataclass, field
from typing import Any, Type

from .environment import Environment, EnvParams
from .types import EnvCarryT


@dataclass
class EnvSpec:
    id: str
    entry_point: str
    kwargs: dict = field(default_factory=dict)


_REGISTRY: dict[str, EnvSpec] = {}


# TODO: add ability to apply wrappers
def register(
    id: str,
    entry_point: str,
    **kwargs,
) -> None:
    if id in _REGISTRY:
        raise ValueError("Environment with such id is already registered. Please choose another one.")

    if not id.startswith("XLand-MiniGrid"):
        if not id.startswith("MiniGrid"):
            raise ValueError("Invalid id format. Should start from XLand-MiniGrid or MiniGrid.")

    env_spec = EnvSpec(id=id, entry_point=entry_point, kwargs=kwargs)
    _REGISTRY[id] = env_spec


def load(name: str) -> Type[Environment]:
    mod_name, env_name = name.split(":")
    mod = importlib.import_module(mod_name)
    env_constructor = getattr(mod, env_name)
    return env_constructor


# TODO: this will give wrong type hints for some envs. How to fix this?
# TODO: for example, Environment[CustomEnvParams, EnvCarry] or Environment[EnvParams, CustomEnvCarry]
def make(id: str, **kwargs: Any) -> tuple[Environment, EnvParams]:
    if id not in _REGISTRY:
        raise ValueError(f"Unregistered environment. Available environments: {', '.join(registered_environments())}")

    env_spec = _REGISTRY[id]
    env_constructor = load(env_spec.entry_point)

    env = env_constructor()
    env_params = env.default_params(**env_spec.kwargs)
    env_params = env_params.replace(**kwargs)

    return env, env_params


def registered_environments() -> tuple[str, ...]:
    return tuple(_REGISTRY.keys())
